# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import uuid
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
import responses
from django.conf import settings
from requests.exceptions import RequestException

from passerelle.apps.qrcode.models import Certificate, QRCodeConnector, Reader
from passerelle.base.signature import check_url
from tests.utils import generic_endpoint_url, setup_access_rights


@pytest.fixture()
def connector(db):
    return setup_access_rights(
        QRCodeConnector.objects.create(
            slug='test',
        )
    )


@pytest.fixture
def mock_wcs():
    settings.KNOWN_SERVICES = {
        'wcs': {
            'default': {
                'title': 'test',
                'url': 'https://wcs.com',
                'secret': 'xxx',
                'orig': 'passerelle',
            },
            'other': {
                'title': 'test',
                'url': 'https://other-wcs.com',
                'secret': 'blah',
                'orig': 'passerelle',
            },
        }
    }


def test_save_campaign(app, connector, freezer):
    freezer.move_to('2022-01-01T10:00:00')
    now = datetime.now(timezone.utc)
    endpoint = generic_endpoint_url('qrcode', 'save-campaign', slug=connector.slug)

    result = app.post_json(
        endpoint,
        params={'metadata': {'toto': 'tata'}},
    )

    campaign_uuid = result.json['data']['uuid']
    campaign = connector.campaigns.get(uuid=campaign_uuid)
    assert campaign
    assert campaign.metadata == {'toto': 'tata'}
    assert campaign.created == now
    assert campaign.modified == now

    freezer.move_to('2022-01-01T10:00:00')
    now = datetime.now(timezone.utc)
    result = app.post_json(
        f'{endpoint}/{campaign_uuid}',
        params={'metadata': {'Georges': 'ravioles'}},
    )

    campaign.refresh_from_db()
    assert campaign.metadata == {'Georges': 'ravioles'}
    assert campaign.modified == now


def test_save_certificate(app, connector):
    endpoint = generic_endpoint_url('qrcode', 'save-certificate', slug=connector.slug)
    campaign = connector.campaigns.create()

    result = app.post_json(
        endpoint,
        params={
            'campaign': str(campaign.uuid),
            'data': {
                'first_name': 'Georges',
                'last_name': 'Abitbol',
            },
            'metadata': {'puissance_intellectuelle': 'BAC +2'},
            'validity_start': '2022-01-01 10:00:00+00:00',
            'validity_end': '2023-01-01 10:00:00+00:00',
        },
    )

    assert result.json['err'] == 0

    certificate_uuid = result.json['data']['uuid']
    assert result.json['data']['qrcode_url'] == f'http://testserver/qrcode/test/get-qrcode/{certificate_uuid}'
    certificate = campaign.certificates.get(uuid=certificate_uuid)

    assert certificate.data['first_name'] == 'Georges'
    assert certificate.data['last_name'] == 'Abitbol'

    assert certificate.metadata['puissance_intellectuelle'] == 'BAC +2'
    assert certificate.validity_start == datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert certificate.validity_end == datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)

    result = app.post_json(
        f'{endpoint}/{certificate_uuid}',
        params={
            'campaign': str(campaign.uuid),
            'data': {
                'first_name': 'Robert',
                'last_name': 'Redford',
            },
            'validity_start': '2024-01-01T10:00:00+00:00',
            'validity_end': '2025-01-01T10:00:00+00:00',
        },
    )

    certificate.refresh_from_db()
    assert certificate.data['first_name'] == 'Robert'
    assert certificate.data['last_name'] == 'Redford'
    assert certificate.validity_start == datetime(2024, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert certificate.validity_end == datetime(2025, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)

    result = app.post_json(
        f'{endpoint}/{certificate_uuid}',
        params={
            'campaign': str(uuid.UUID('00000000-0000-0000-0000-000000000000')),
        },
        status=400,
    )

    assert result.json['err'] == 1
    assert result.json['err_desc'] == "You can't change the campaign of an existing certificate"


def test_get_certificate(app, connector):
    campaign = connector.campaigns.create()
    certificate = campaign.certificates.create(
        data={
            'first_name': 'Georges',
            'last_name': 'Abitbol',
        },
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )

    endpoint = generic_endpoint_url('qrcode', 'get-certificate', slug=connector.slug)
    result = app.get(f'{endpoint}/{certificate.uuid}')

    assert result.json == {
        'err': 0,
        'data': {
            'uuid': str(certificate.uuid),
            'data': {'first_name': 'Georges', 'last_name': 'Abitbol'},
            'validity_start': '2022-01-01T10:00:00+00:00',
            'validity_end': '2023-01-01T10:00:00+00:00',
            'qrcode_url': f'http://testserver/qrcode/test/get-qrcode/{certificate.uuid}',
        },
    }


def test_get_qrcode(app, connector):
    campaign = connector.campaigns.create(
        key='5e8176e50d45b67e9db875d6006edf3ba805ff4ef4d945327012db4c797be1be'
    )
    certificate = campaign.certificates.create(
        uuid=uuid.UUID('12345678-1234-5678-1234-567812345678'),
        data={
            'first_name': 'Georges',
            'last_name': 'Abitbol',
        },
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )
    endpoint = generic_endpoint_url('qrcode', 'get-qrcode', slug=connector.slug)

    response = app.get(f'{endpoint}/{certificate.uuid}')
    assert response.headers['Content-Type'] == 'image/png'
    with open('tests/data/qrcode/test-qrcode.png', 'rb') as expected_qrcode:
        # just check images are the same. Decoded content is tested javascript-side.
        assert response.body == expected_qrcode.read()


def test_save_reader(app, connector):
    endpoint = generic_endpoint_url('qrcode', 'save-reader', slug=connector.slug)
    campaign = connector.campaigns.create()

    result = app.post_json(
        endpoint,
        params={
            'campaign': str(campaign.uuid),
            'validity_start': '2022-01-01 10:00:00+00:00',
            'validity_end': '2023-01-01 10:00:00+00:00',
            'readable_metadatas': 'name,last_name',
            'metadata': {'toto': 'tata'},
        },
    )

    assert result.json['err'] == 0

    reader_uuid = result.json['data']['uuid']
    assert result.json['data']['url'] == f'http://testserver/qrcode/test/open-reader/{reader_uuid}'
    reader = campaign.readers.get(uuid=reader_uuid)

    assert reader.readable_metadatas == 'name,last_name'
    assert reader.validity_start == datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert reader.validity_end == datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert reader.metadata == {'toto': 'tata'}

    result = app.post_json(
        f'{endpoint}/{reader_uuid}',
        params={
            'validity_start': '2024-01-01T10:00:00+00:00',
            'validity_end': '2025-01-01T10:00:00+00:00',
            'metadata': {'Georges': 'ravioles'},
        },
    )

    reader.refresh_from_db()
    assert reader.validity_start == datetime(2024, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert reader.validity_end == datetime(2025, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc)
    assert reader.metadata == {'Georges': 'ravioles'}


def test_get_reader(app, connector):
    campaign = connector.campaigns.create()
    reader = campaign.readers.create(
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )

    endpoint = generic_endpoint_url('qrcode', 'get-reader', slug=connector.slug)
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert result.json == {
        'err': 0,
        'data': {
            'uuid': str(reader.uuid),
            'validity_start': '2022-01-01T10:00:00+00:00',
            'validity_end': '2023-01-01T10:00:00+00:00',
            'url': f'http://testserver/qrcode/test/open-reader/{reader.uuid}',
        },
    }


def test_open_reader(app, connector, freezer):
    campaign = connector.campaigns.create()
    reader = campaign.readers.create(
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )

    endpoint = generic_endpoint_url('qrcode', 'open-reader', slug=connector.slug)
    freezer.move_to('2022-01-01T09:59:59')
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert 'Reader isn\'t usable yet' in result.body.decode('utf-8')

    freezer.move_to('2022-01-01T10:00:00')
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert result.pyquery(f'qrcode-reader[verify-key="{campaign.hex_verify_key}"]')
    assert not result.pyquery('qrcode-reader[tally-url]')

    campaign.metadata = {'tally': True}
    campaign.save()
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert result.pyquery(f'qrcode-reader[verify-key="{campaign.hex_verify_key}"][tally-url]')

    campaign.metadata = {'tally': False}
    campaign.save()

    reader.metadata = {'tally': True}
    reader.save()
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert result.pyquery(f'qrcode-reader[verify-key="{campaign.hex_verify_key}"][tally-url]')

    freezer.move_to('2023-01-01T10:00:01')
    result = app.get(f'{endpoint}/{reader.uuid}')

    assert 'Reader has expired.' in result.body.decode('utf-8')


def test_read_metadata(app, connector, freezer):
    campaign = connector.campaigns.create()
    reader = campaign.readers.create(
        readable_metadatas='first_name,puissance_intellectuelle',
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )

    certificate = campaign.certificates.create(
        uuid=uuid.UUID('12345678-1234-5678-1234-567812345678'),
        metadata={'first_name': 'Georges', 'last_name': 'Abitbol', 'puissance_intellectuelle': 'BAC +2'},
        validity_start=datetime(2022, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
        validity_end=datetime(2023, 1, 1, 10, 0, 0, 0, tzinfo=timezone.utc),
    )

    endpoint = generic_endpoint_url('qrcode', 'read-metadata', slug=connector.slug)
    freezer.move_to('2022-01-01T09:59:59')
    result = app.get(f'{endpoint}/{reader.uuid}', params={'certificate': certificate.uuid})

    assert result.json['err'] == 1
    assert result.json['err_desc'] == 'Reader isn\'t usable yet.'

    freezer.move_to('2022-01-01T10:00:00')
    result = app.get(f'{endpoint}/{reader.uuid}', params={'certificate': certificate.uuid})

    assert result.json['err'] == 0
    assert result.json['data'] == {'first_name': 'Georges', 'puissance_intellectuelle': 'BAC +2'}

    freezer.move_to('2023-01-01T10:00:01')
    result = app.get(f'{endpoint}/{reader.uuid}', params={'certificate': certificate.uuid})

    assert result.json['err'] == 1
    assert result.json['err_desc'] == 'Reader has expired.'


def test_tally(app, connector, freezer):
    freezer.move_to('2022-01-01T09:59:59')
    endpoint = generic_endpoint_url('qrcode', 'tally', slug=connector.slug)

    campaign = connector.campaigns.create(metadata={'tally': True})
    reader = campaign.readers.create()
    certificate = campaign.certificates.create()

    def tally(*certificates, since=0, event_uuid=None, **params):
        now = datetime.now(timezone.utc)

        result = app.post_json(
            f'{endpoint}/{reader.uuid}',
            params={
                'since': int(datetime.timestamp(since)) if since != 0 else 0,
                'events': [
                    {
                        'uuid': event_uuid or str(uuid.uuid4()),
                        'certificate': str(certificate),
                        'timestamp': int(datetime.timestamp(now)),
                    }
                    for certificate in certificates
                ],
            }
            | params,
        )

        assert result.json['err'] == 0
        assert result.json['data']['timestamp'] == int(datetime.timestamp(now))
        return result.json['data']['certificates']

    result = tally()
    assert result[str(certificate.uuid)]['credit'] == 0

    result = tally(certificate.uuid)
    assert result[str(certificate.uuid)]['credit'] == -1

    # No credit: event is not saved
    events = list(certificate.events.all())
    assert len(events) == 0

    # Check invalid certificates are ignored
    result = tally('deadbeef-0000-0000-0000-000000000000')
    assert result[str(certificate.uuid)]['credit'] == 0
    assert 'deadbeef-0000-0000-0000-000000000000' not in result

    now = datetime.now(timezone.utc)

    # Credit certificate and tally it
    certificate.events.create(happened=now, metadata={'credit': 1})
    result = tally(certificate.uuid)
    events = list(certificate.events.all())

    assert len(events) == 2
    event = events[1]
    assert result[str(certificate.uuid)]['credit'] == 0
    assert event.reader == reader
    assert event.happened == now
    assert event.received == now
    assert event.metadata == {'credit': -1}

    # Check duplicate events are ignored
    result = tally(certificate.uuid, event_uuid=str(event.uuid))
    events = list(certificate.events.all())
    assert len(events) == 2
    assert events[1] == event

    # Check since filtering
    yesterday = now
    freezer.tick(timedelta(days=1))

    assert str(certificate.uuid) in tally(since=yesterday + timedelta(minutes=1))
    assert str(certificate.uuid) not in tally(since=yesterday + timedelta(minutes=2))

    # specifying a certificate should return it even if filtered by since
    assert str(certificate.uuid) in tally(certificate.uuid, since=yesterday + timedelta(minutes=2))

    # Asynchronous invalid tallying should save overdraft events
    result = tally(certificate.uuid, asynchronous=True)
    assert result[str(certificate.uuid)]['credit'] == -1
    events = list(certificate.events.all())
    assert len(events) == 3
    assert events[2].metadata['credit'] == -1

    # expired certificates should be ignored
    expired_certificate = campaign.certificates.create(
        validity_end=datetime.now(timezone.utc) - timedelta(seconds=1)
    )
    not_expired_certificate = campaign.certificates.create(validity_end=datetime.now(timezone.utc))
    assert str(expired_certificate.uuid) not in tally()
    assert str(not_expired_certificate.uuid) in tally()


@mock.patch('passerelle.base.models.ProxyLogger.error')
@responses.activate
def test_tally_trigger_wcs(mock_log_error, app, connector, freezer, mock_wcs):
    freezer.move_to('2023-01-01T10:00:00')
    endpoint = generic_endpoint_url('qrcode', 'tally', slug=connector.slug)

    campaign = connector.campaigns.create(metadata={'tally': True})
    reader = campaign.readers.create()
    certificate = campaign.certificates.create()

    def tally():
        now = datetime.now(timezone.utc)
        certificate.events.create(happened=now, metadata={'credit': 1})

        result = app.post_json(
            f'{endpoint}/{reader.uuid}',
            params={
                'since': 0,
                'events': [
                    {
                        'uuid': str(uuid.uuid4()),
                        'certificate': str(certificate.uuid),
                        'timestamp': int(datetime.timestamp(now)),
                    }
                ],
            },
        )

        assert result.json['err'] == 0
        return result.json['data']['certificates']

    trigger_url = 'https://wcs.com/trigger/blah'
    certificate.metadata = {'trigger_url': trigger_url}
    certificate.save()

    responses.post(trigger_url)
    result = tally()
    assert str(certificate.uuid) in result

    assert len(responses.calls) == 1
    assert responses.calls[0].response.url.startswith(trigger_url)
    assert check_url(responses.calls[0].response.url, settings.KNOWN_SERVICES['wcs']['default']['secret'])

    responses.post('https://wcs.com/trigger/blah', body=RequestException('kaboume'))
    result = tally()

    # certificate should be returned anyway in case of trigger error
    assert str(certificate.uuid) in result
    mock_log_error.assert_called_with(
        'Error while triggering w.c.s. for certificate %s : %s', certificate.uuid, mock.ANY
    )

    certificate.metadata = {'trigger_url': 'https://invalid-wcs-url.com/trigger/blah'}
    certificate.save()
    result = tally()

    # certificate should be returned anyway in case of trigger error
    assert str(certificate.uuid) in result
    mock_log_error.assert_called_with(
        "Can't find a suitable configured WCS service for url %s", 'https://invalid-wcs-url.com/trigger/blah'
    )


def test_add_event(app, connector, freezer):
    freezer.move_to('2023-01-01T10:00:01')
    campaign = connector.campaigns.create(metadata={'tally': True})
    certificate = campaign.certificates.create()
    now = datetime.now(timezone.utc)

    endpoint = generic_endpoint_url('qrcode', 'add-event', slug=connector.slug)
    app.post_json(
        endpoint,
        params={
            'certificate': str(certificate.uuid),
            'metadata': {'toto': 'tata'},
        },
    )

    certificate.refresh_from_db()
    events = list(certificate.events.all())
    assert len(events) == 1
    assert events[0].metadata == {'toto': 'tata'}
    assert events[0].received == now
    assert events[0].happened == now


def test_get_events(app, connector, freezer):
    freezer.move_to('2023-01-01T10:00:00')
    now = datetime.now(timezone.utc)

    swimming_pool = connector.campaigns.create()
    sausage_fair = connector.campaigns.create()

    john = swimming_pool.certificates.create()
    jack = sausage_fair.certificates.create()

    swimming_pool_entry = swimming_pool.readers.create()
    sausage_fair_entry = sausage_fair.readers.create()

    john_buys_swimming_pool_tickets_event = john.events.create(happened=now, metadata={'credit': 10})

    freezer.move_to('2023-02-01T10:00:00')
    now = datetime.now(timezone.utc)
    john_goes_to_swimming_pool_event = john.events.create(
        reader=swimming_pool_entry, happened=now, metadata={'credit': -1}
    )
    jack_goes_to_sausage_fair_event = jack.events.create(
        reader=sausage_fair_entry, happened=now, metadata={'credit': -1}
    )

    john_buys_swimming_pool_tickets = {
        'uuid': str(john_buys_swimming_pool_tickets_event.uuid),
        'certificate': str(john.uuid),
        'reader': None,
        'received': '2023-01-01T10:00:00+00:00',
        'happened': '2023-01-01T10:00:00+00:00',
        'metadata': {'credit': 10},
    }

    john_goes_to_swimming_pool = {
        'uuid': str(john_goes_to_swimming_pool_event.uuid),
        'certificate': str(john.uuid),
        'reader': str(swimming_pool_entry.uuid),
        'received': '2023-02-01T10:00:00+00:00',
        'happened': '2023-02-01T10:00:00+00:00',
        'metadata': {'credit': -1},
    }

    jack_goes_to_sausage_fair = {
        'uuid': str(jack_goes_to_sausage_fair_event.uuid),
        'certificate': str(jack.uuid),
        'reader': str(sausage_fair_entry.uuid),
        'received': '2023-02-01T10:00:00+00:00',
        'happened': '2023-02-01T10:00:00+00:00',
        'metadata': {'credit': -1},
    }

    endpoint = generic_endpoint_url('qrcode', 'events', slug=connector.slug)

    result = app.get(endpoint)
    assert result.json == {
        'err': 0,
        'data': [
            john_buys_swimming_pool_tickets,
            john_goes_to_swimming_pool,
            jack_goes_to_sausage_fair,
        ],
    }

    result = app.get(f'{endpoint}?campaign={swimming_pool.uuid}')
    assert result.json == {
        'err': 0,
        'data': [
            john_buys_swimming_pool_tickets,
            john_goes_to_swimming_pool,
        ],
    }

    result = app.get(f'{endpoint}?certificate={john.uuid}')
    assert result.json == {
        'err': 0,
        'data': [
            john_buys_swimming_pool_tickets,
            john_goes_to_swimming_pool,
        ],
    }

    result = app.get(f'{endpoint}?reader={swimming_pool_entry.uuid}')
    assert result.json == {
        'err': 0,
        'data': [john_goes_to_swimming_pool],
    }

    result = app.get(f'{endpoint}?since=2023-02-01T10:00:00%2B00:00')
    assert result.json == {
        'err': 0,
        'data': [john_goes_to_swimming_pool, jack_goes_to_sausage_fair],
    }

    result = app.get(f'{endpoint}?until=2023-01-01T10:00:00%2B00:00')
    assert result.json == {
        'err': 0,
        'data': [
            john_buys_swimming_pool_tickets,
        ],
    }

    result = app.get(f'{endpoint}?tally=True')
    assert result.json == {
        'err': 0,
        'data': [john_goes_to_swimming_pool, jack_goes_to_sausage_fair],
    }

    result = app.get(f'{endpoint}?tally=False')
    assert result.json == {
        'err': 0,
        'data': [
            john_buys_swimming_pool_tickets,
        ],
    }


MISSING = object()


@pytest.mark.parametrize('value', [MISSING, None, ''], ids=['missing', 'null', 'empty string'])
class TestOptional:
    def test_certificate_validity_start(self, value, app, connector):
        campaign = connector.campaigns.create()
        params = {
            'campaign': str(campaign.uuid),
            'data': {
                'first_name': 'Georges',
                'last_name': 'Abitbol',
            },
            'validity_end': '2023-01-01 10:00:00+00:00',
        }
        if value is not MISSING:
            params['validity_start'] = value

        app.post_json('/qrcode/test/save-certificate/', params=params)
        assert Certificate.objects.get().validity_start is None

    def test_certificate_validity_end(self, value, app, connector):
        campaign = connector.campaigns.create()
        params = {
            'campaign': str(campaign.uuid),
            'data': {
                'first_name': 'Georges',
                'last_name': 'Abitbol',
            },
            'validity_start': '2023-01-01 10:00:00+00:00',
        }
        if value is not MISSING:
            params['validity_end'] = value

        app.post_json('/qrcode/test/save-certificate/', params=params)
        assert Certificate.objects.get().validity_end is None

    def test_reader_validity_start(self, value, app, connector):
        campaign = connector.campaigns.create()
        params = {
            'campaign': str(campaign.uuid),
            'validity_end': '2023-01-01 10:00:00+00:00',
        }
        if value is not MISSING:
            params['validity_start'] = value

        app.post_json('/qrcode/test/save-reader/', params=params)
        assert Reader.objects.get().validity_start is None

    def test_reader_validity_end(self, value, app, connector):
        campaign = connector.campaigns.create()
        params = {
            'campaign': str(campaign.uuid),
            'validity_start': '2023-01-01 10:00:00+00:00',
        }
        if value is not MISSING:
            params['validity_end'] = value

        app.post_json('/qrcode/test/save-reader/', params=params)
        assert Reader.objects.get().validity_end is None
